
/*
   pfodUnoIRTemp
   Control your IR Temperature Meter from your Android mobile

   see www.pfod.com.au for other projects

   (c)2013 Forward Computing and Control Pty. Ltd.
   This code may be freely used for both private and commerical use.
   Provide this copyright is maintained.

   NOTE: This sketch uses Interrupt 1
   and pins D2,D3,D4

*/

#include <util/atomic.h>  // from AVR-Lib to ensure when we clear the IR Temp data, interrupts don't interrupt it.
#include "pfodEEPROM.h"
#include "pfodParser.h"

//#define IRTERMP_DEBUG // uncomment this line to get lots more output

pfodParser parser;   // define a parser

// some version of Arduino IDE need these methods declared here.
void setupIRTemp (byte pinAcquire, byte pinClock, byte pinData);
void sensorEnable(bool state);
bool validData(volatile byte data[]);
void startAnalogRead();
int processAnalogReading();
int readIntTemperature(volatile byte data[]);
float readTemperature(volatile byte data[]);
float readTemperature(volatile byte data[]);
void resetData(void);


// define a bit of AVR-Lib C magic to get the equivalent of assembler sbi
#ifndef sbi
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#endif


static const byte PIN_DATA    = 2;
static const byte PIN_CLOCK   = 3;  // Must be either pin 2 or pin 3
static const byte PIN_ACQUIRE = 4;
static const long IRTEMP_TIMEOUT = 1600;  // milliseconds, if IR temp reading takes more than this mS then fail
//static const long IRTEMP_DATA_AMBIENT = 0x66;  not used
static const long IRTEMP_DATA_IR =      0x4C;

boolean SCALE = false; // Celsius: false, Fahrenheit: true
float  irTemperature;
float averageTemp;
float lastTemp;

#define MAX_INITIAL_NO_SAMPLES 5
#define MAX_AVERAGE_BUFFER_SIZE (2<<MAX_INITIAL_NO_SAMPLES)
byte initialNoSamples = 2;  // 2== 8, zero based 2 4 8 16 32 are the choices MUST be < MAX_INITIAL_NO_SAMPLES
int numberOfAverages = 2 << initialNoSamples;
// numberOfAverages MUST be < MAX_AVERAGE_BUFFER_SIZE

int supplyInput = A0;    // select the input pin for the potentiometer
int supplyValue = 0;  // supply volts count
float rechargeVoltsLimit = 3.00; // = 774 counts in my circuit
float volts; // current volts
boolean lowVolts = false; // set to true if supply volts < 3.00

int irReading = 0;  // the ir temp count
long IRTempStartTime = 0; // time when last irTemp reading started
long rawDataTime = 0;
long rawDataStartTime = 0;
long time = 0;
long sampleTime = 0;
void setup(void) {
  Serial.begin(9600); //Set BluetoothFrame BaudRate to default baud rate 38400 set baud rate for Sparkfun Silver to 115200
  delay(1000);
  // setupBlueToothConnection();
  // Serial.println(F("IRTemp Meter for Android"));
  // Serial.println(F("~~~~~~~~~~~~~~"));
  parser.connect(&Serial);
  analogReference(INTERNAL); // 1.1V or 2.56V depending on board
  // adjust resistor divider to suit, here set for 1.1V
  analogRead(supplyInput); // do one read here to setup the ADC using the library code
  startAnalogRead(); // start the first reading of the main loop
  clearAverage(irReading); // set all to 0's
  time = millis();
  sampleTime = time;
  rawDataStartTime = time;
  rawDataTime = time;

  setupIRTemp(PIN_ACQUIRE, PIN_CLOCK, PIN_DATA);
}

void setupBlueToothConnection() {
  sendBlueToothCommand("+STWMOD=0");
  sendBlueToothCommand("+STAUTO=0");
  sendBlueToothCommand("+STOAUT=1");
  delay(3000); // This delay is required.
  sendBlueToothCommand("+INQ=1");
  delay(3000); // This delay is required.
}

static const char* BT_Name = "IRTempBluetooth";
/**
   this method only needs to be called once to initialize the Film bluetooth module
   after that the above setupBlueToothConnection() needs to be called on each power up.
   it keeps the defalt pin of 1234
*/
/**
  void setupBlueToothConnection() {
    sendBlueToothCommand("+STWMOD=0");
    sendBlueToothCommand("+STNA=IRTempBluetooth");
    sendBlueToothCommand("+STAUTO=0");
    sendBlueToothCommand("+STOAUT=1");
    sendBlueToothCommand("+STPIN=1234");
    delay(3000); // This delay is required.
    sendBlueToothCommand("+INQ=1");
    delay(3000); // This delay is required.
  }
*/

//Send the command value to Bluetooth Frame
void sendBlueToothCommand(char command[])
{

  Serial.println();
  delay(200);
  Serial.print(command);
  delay(200);
  Serial.println();
  delay(1000);
  Serial.flush();
}
void loop(void) {
  // get temp once per sec
  long t = millis();
  if (((t - (sampleTime + 1000)) > 0) && (IRTempStartTime == 0)) {
    while ((t - (sampleTime + 1000)) > 0) {
      sampleTime += 1000; // alway increment by 1000 this prevents drift accumulation
    }
    startAnalogRead(); // start a voltage reading
    startGetTemp(); // sets IRTempStartTime
#ifdef IRTERMP_DEBUG
    Serial.println(F("start capture"));
    Serial.print(F("time = "));
    Serial.println(sampleTime - time);
#endif
  }
  int tempReading = getTemp(); // clears IRTempStartTime is sample complete or if measurement times out.
  if (tempReading != 0) {
    irReading = tempReading; // use this when resetting average filter
    lastTemp =  irIntToDeg((float)tempReading);
    averageTemp = average(tempReading);
    rawDataTime = t;
    Serial.print(((float)(rawDataTime - rawDataStartTime)) / 1000.0);
    Serial.print(F("Sec,"));
    printTemperature(lastTemp);  // raw data
    Serial.println();

#ifdef IRTERMP_DEBUG
    Serial.print(F("last temp "));
    Serial.print(lastTemp);
    Serial.print(F(" average Temp "));
    Serial.print(averageTemp);
    Serial.println();
#endif

  }

  supplyValue =   processAnalogReading(); // handle any new reading, return -1 if no reading yet
  if (supplyValue >= 0) {
    // have a reading
    // 848 = 3.287V
    volts = supplyValue / 848.0 * 3.287;
    if (volts < 3.00) {
      lowVolts = true;
    }
    else {
      lowVolts = false;
    }
#ifdef IRTERMP_DEBUG
    Serial.print(F(" Analog reading "));
    Serial.print(volts);
    Serial.println(F("V"));
#endif
  }
  processSerialInput();
}

void printTemperature( float  temperature) {
  if (isnan(temperature)) {
    Serial.print(F("Failed"));
  }
  else {
    Serial.print(temperature);
    Serial.print(SCALE  ?  " F"  :  " C");
  }
}

void switchScale() {
  SCALE = !SCALE;
}

//  float irTemperature = irTemp.getIRTemperature(SCALE);

// send the main menu based on last Temp and average
void sendMainMenu() {
  //irReading = irReading+160;
  //lastTemp = irIntToDegC((float)irReading);
  //averageTemp = average(irReading);

  float temperature = lastTemp;
  Serial.print(F("{.IR Temperature Sensor\nPress IR Temp to zero sample time\nPress Average IR Temp to reset it`1000|"));
  Serial.print(F("t~IR Temperature\n"));
  printTemperature(temperature);
  Serial.print(F("|"));
  Serial.print(F("a~Average IR Temp\n"));
  temperature = averageTemp;
  printTemperature(temperature);
  Serial.print(F(""));
  /*
    Serial.print(F("|v~Battery\n"));
    if (lowVolts) {
    Serial.print(F("Needs Charging\n"));
    }
    Serial.print(volts);
    Serial.print(F("V"));
  */
  Serial.print(F("|"));
  Serial.print(F("c~Switch to\n"));
  if (SCALE) {
    Serial.print(F("Centigrade (C)"));
  } else {
    Serial.print(F("Fahrenheit (F)"));
  }
  Serial.print(F("|"));
  Serial.print(F("s~Set number readings averaged"));
  Serial.print(F("}"));
}

// send the settings menu
void sendSettingsMenu() {
  Serial.print(F("{?m`"));
  Serial.print(initialNoSamples);
  Serial.print(F("~Set number of samples to be averaged"));
  Serial.print(F("|Average 2 Readings|Average 4 Readings|Average 8 Readings|Average 16 Readings|Average 32 Readings"));
  Serial.print(F("}"));
}

float convertFahrenheit(float celsius) {
  return (celsius * 9 / 5 + 32);
}

void processSerialInput() { // run over and over
  // process IR reading
  byte cmd = parser.parse();
  if (cmd != 0) {
#ifdef IRTERMP_DEBUG
    Serial.print(F("Cmd:"));
    Serial.write(cmd);
    Serial.println();
#endif
    if (cmd == ((byte)'c')) {
      switchScale();
      sendEmptyResponse();
    } else if (cmd == (byte)'.') {
      // send main menu
      sendMainMenu();
    } else if (cmd == (byte)'t') {
      // reset sample time to 0
      rawDataStartTime = millis();
      rawDataTime = rawDataStartTime;
      sendEmptyResponse(); // cmd processed
    } else if (cmd == (byte)'a') {
      // clear the average to current temp
      clearAverage(irReading);
      sendEmptyResponse(); // cmd processed
    } else if (cmd == (byte)'s') {
      // clear the average to current temp
      sendSettingsMenu();
    } else if (cmd == (byte)'m') {
      byte* idxPtr = parser.getFirstArg();
      // parse 1 arg as a number
      long longResult;
      parser.parseLong(idxPtr, &longResult); // only one arg so no need to loop
      initialNoSamples = (byte)longResult;
      setNoReadingsToAverage((int)initialNoSamples);
      sendEmptyResponse(); // cmd processed
    } else {
      // ignore the rest
      sendEmptyResponse();
    }
    // else keep looping
  }
}

void sendEmptyResponse() {
  Serial.println(F("{}"));
}
// set the new number of reading to index and
// re-initialize the average
void setNoReadingsToAverage(int indexSelected) {
  numberOfAverages = 2 << indexSelected;
#ifdef IRTERMP_DEBUG
  Serial.print(F("Set no readings to average to "));
  Serial.println(numberOfAverages);
#endif
  clearAverage(irReading);
}

// don't use floats here as errors will accumulate over time.
// max size of buffer is 32
// int reading for 25degC is 4770
// int reading for 200degC is 7568
// int is 65535 i.e. ~8.6 times 200degC
// so need to use long for average so it does not over flow
struct average_buffer_struct {
  unsigned int buffer[MAX_AVERAGE_BUFFER_SIZE];
  unsigned int idx;
  long average;
}
average_buffer;


// add a new reading and return the new average deg C
float average(int nextReading) {
  if (average_buffer.buffer[0] == 0) {
    // fill with this reading to start with
    return clearAverage(nextReading);
  }
  // else
  int i = (unsigned int)(average_buffer.idx + 1) % numberOfAverages;
  average_buffer.idx = i;
#ifdef IRTERMP_DEBUG
  Serial.print(F(" i = "));
  Serial.print(i);
  // sub the current value, add the new value and divide by AVERAGE_BUFFER_SIZE
  Serial.print(F(" avg = "));
  Serial.print(average_buffer.average);
  Serial.print(F(" buf[i] = "));
  Serial.print(average_buffer.buffer[i]);
  Serial.print(F(" nextReading = "));
  Serial.print(nextReading);
#endif
  average_buffer.average = average_buffer.average - average_buffer.buffer[i];
  average_buffer.buffer[i] = (long)nextReading;
  average_buffer.average = average_buffer.average + nextReading;
#ifdef IRTERMP_DEBUG
  Serial.print(F(" new avg = "));
  Serial.print(average_buffer.average);
  Serial.print(F("  AvgTmp = "));
  Serial.println(average_buffer.average / numberOfAverages);
#endif
  return  irIntToDeg(((float)average_buffer.average) / numberOfAverages);
}

// this method lets the user clear out the old averages
float clearAverage(int startingReading) {
#ifdef IRTERMP_DEBUG
  Serial.print(F("startingReading ="));
  Serial.println(startingReading);
#endif

  for (int i = 0; i < numberOfAverages; i++) {
    average_buffer.buffer[i] = startingReading;
  }
  average_buffer.idx = 0;
  average_buffer.average = ((long)startingReading) * numberOfAverages;
#ifdef IRTERMP_DEBUG
  Serial.print(F("Cleared AvgTemp="));
  Serial.println(average_buffer.average / numberOfAverages);
#endif
  return  irIntToDeg( ((float)average_buffer.average) / numberOfAverages);
}

float irIntToDeg(float reading) {
  float result = reading / 16.0 - 273.15;
  if (SCALE) {
    result = convertFahrenheit(result);
  }
  return result;
}

void startGetTemp() {
  if (IRTempStartTime == 0) {
    IRTempStartTime = millis();
    resetData();
    sensorEnable(true);
  }
}

int getTemp() {
  if (IRTempStartTime == 0) {
    return 0; // not started or have finished
  }
  else {
    long t = millis();
    if ( (IRTempStartTime > 0) &&  ((t - IRTempStartTime) > IRTEMP_TIMEOUT) ) {
      sensorEnable(false);
#ifdef IRTERMP_DEBUG
      Serial.println(F("failed"));
      Serial.print(F("time = "));
      Serial.println(t - time);
#endif
      IRTempStartTime = 0;
      return 0;
    }
    else {
      int irTemperature = getIRTemperature();
      if (irTemperature != 0) {
#ifdef IRTERMP_DEBUG
        Serial.print(F("time = "));
        Serial.println(t - time);
        printTemperature(irTemperature);
#endif
        IRTempStartTime = 0;
      }
      return irTemperature;
    }
  }
}

void printTemperature(
  int  temperature) {
  float temp;
  Serial.print(F("IR temperature: "));

  if (temperature == 0) {
    Serial.println("Failed");
  }
  else {
    temp = (temperature / 16.0 - 273.15);
    if (SCALE) temp = convertFahrenheit(temp);

    Serial.print(temp);
    Serial.println(SCALE  ?  " F"  :  " C");
  }

}


extern void irTempInterrupt(void);

static const byte IRTEMP_DATA_SIZE = 10;

volatile byte irTempData[IRTEMP_DATA_SIZE];
volatile byte irTempDataBit;
volatile byte irTempDataIndex;
volatile bool readingStarted;
volatile bool irTempSensorEnabled;

byte irTempPinAcquire;
byte irTempPinClock;
byte irTempPinData;

void setupIRTemp (
  byte pinAcquire,
  byte pinClock,
  byte pinData) {

  irTempPinAcquire = pinAcquire;
  irTempPinClock =   pinClock;
  irTempPinData =    pinData;

  pinMode(irTempPinAcquire, OUTPUT);
  pinMode(irTempPinClock,   INPUT);
  pinMode(irTempPinData,    INPUT);

  digitalWrite(irTempPinAcquire, HIGH);
  digitalWrite(irTempPinClock,   HIGH);
  digitalWrite(irTempPinData,    HIGH);
  sensorEnable(false);
  byte interruptNumber = (pinClock == 2)  ?  0  :  1;
  attachInterrupt(interruptNumber, irTempInterrupt, FALLING);
}



int getIRTemperature(void) {  // note use int here as int varies from board to board
  byte dataType = IRTEMP_DATA_IR;
  int temperature = 0;
  long  timeStart = millis();

  if (irTempDataIndex < IRTEMP_DATA_SIZE) {
    return 0;
  }

  if (irTempData[0] == dataType) {
    if (validData(irTempData)) temperature = readIntTemperature(irTempData);
  }
  else {
    if (validData(irTempData + 5)) temperature = readIntTemperature(irTempData + 5);
  }
#ifdef IRTERMP_DEBUG
  Serial.println(F("got full data"));
#endif
  sensorEnable(false);
  return (temperature);
}

int readIntTemperature(
  volatile byte data[]) { // need data volatile since it is updated via an interrpt

  int msb = data[1] << 8;
  int lsb = data[2];

  return (msb + lsb);
}

float readTemperature(
  volatile byte data[]) {

  int msb = data[1] << 8;
  int lsb = data[2];

  return ((msb + lsb) / 16.0 - 273.15);
}

void resetData(void) {
  ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
    for (byte i = 0;  i < IRTEMP_DATA_SIZE;  i ++) irTempData[i] = 0;

    irTempDataBit = 8;
    irTempDataIndex = 0;
  }
}

void sensorEnable(
  bool state) {
  irTempSensorEnabled = state;
  digitalWrite(irTempPinAcquire, ! irTempSensorEnabled);
}

// receive 5 data bytes of 8 bits each
void irTempInterrupt(void) {
  if (irTempSensorEnabled) {
    if (irTempDataIndex < IRTEMP_DATA_SIZE) {
      irTempDataBit --;
      irTempData[irTempDataIndex] |=
        digitalRead(irTempPinData) << irTempDataBit;

      if (irTempDataBit == 0) {
        irTempDataBit = 8;
        irTempDataIndex ++;
      }
    }
  }
}

bool validData(
  volatile byte data[]) {
  byte checksum = (data[0] + data[1] + data[2]) & 0xff;
  return (data[3] == checksum  &&  data[4] == 0x0d);
}


/**
    use void analogReference(uint8_t mode) to set reference
   and then use int analogRead(uint8_t pin) once to set the ADC pin to read
   also does on initial read
   Then use the method startAnalogRead() to start the conversion
   then continue processing.
   Call processAnalogReading to process the reading.
   Note this processAnalogReading will just return if conversion has not completed
*/
boolean analogReadStarted = false;
void startAnalogRead() {
  if (analogReadStarted) {
    return;
  }
#if defined(ADCSRA) && defined(ADCL)
  // start the conversion
  sbi(ADCSRA, ADSC);
  analogReadStarted = true;
#endif // else don't have ADC
}

/**
   Call this from loop() to process reading when it is available
*/
int processAnalogReading() {
  uint8_t low, high;
  int reading;

#if defined(ADCSRA) && defined(ADCL)
  // ADSC is cleared when the conversion finishes
  if ((!analogReadStarted) || (bit_is_set(ADCSRA, ADSC))) {
    return -1; // return if still converting
  }
  analogReadStarted = false;
  // we have to read ADCL first; doing so locks both ADCL
  // and ADCH until ADCH is read.  reading ADCL second would
  // cause the results of each conversion to be discarded,
  // as ADCL and ADCH would be locked when it completed.
  low  = ADCL;
  high = ADCH;
#else
  // we dont have an ADC, return 0
  low  = 0;
  high = 0;
#endif
  // combine the two bytes
  reading = (high << 8) | low;
  // do what you need to with the reading here.
  return reading;
}




