/**
 * File:    pfodIRTemp.cpp
 * (c)2013 Forward Computing and Control Pty. Ltd
 * see www.pfod.com.au
 *
 * NOTE: No argument constructor uses Interrupt 1
 * and pins D2(data),D3(clock),D4(trigger)
 *
 * To start a measurement call 
 * triggerSensor() 
 * triggerSensor can be called every loop as subsequent calls do nothing until a non-reading has been returned by 
 * getIRtemperature();
 * 
 * Calls to getIRtemperature() will return pfodIRTemp.NO_DATA (0) until a reading is available
 * When a reading is available getIRtemperature() will return either the raw reading (16*Kelvin)
 * or < pfodIRTemp.NO_DATA if error (pfodIRTemp.INVALID_DATA (-1) or pfodIRTemp.READING_TIMEOUT (-2))
 *
 * Sample code
   #include "pfodIRTemp.h" 
   pfodIRTemp irTemp;  // construct class to handle sensor
   
   bool fahrenheit = false; // set to true for Fahrenheit instead of Celsius
   void setup() {
     Serial.begin(9600);
   }
   
   void loop() {
     irTemp.triggerSensor();
     int reading = irTemp.getIRTemperature();
     if (reading != pfodIRTemp::NO_DATA) {
       // have a reading
       if (reading <  pfodIRTemp::NO_DATA) {
          Serial.println(F(" ,Invalid Reading"));
       } else {
          if (fahrenheit) {
            Serial.print(irTemp.convertToFahrenheit(reading)); 
            Serial.println(F(" ,F"));
          } else {  
            Serial.print(irTemp.convertToCelsius(reading));
            Serial.println(F(" ,C"));
          }  
       }
     }  
   } 
 *
 */

#include "pfodIRTemp.h"
// class variable initialized.  These are class static variables
 const byte pfodIRTemp::pfodIRTemp::IRTEMP_DATA_SIZE = 10; // this needs to be above reference below

 const long pfodIRTemp::IRTEMP_TIMEOUT = 1000;  // milliseconds reading timeout
 const long pfodIRTemp::IRTEMP_DATA_IR = 0x4C;
 
 const int pfodIRTemp::NO_DATA = 0;
 const int pfodIRTemp::INVALID_DATA = -1;
 const int pfodIRTemp::READING_TIMEOUT = -2;  
 
 byte pfodIRTemp::PIN_ACQUIRE;
 byte pfodIRTemp::PIN_DATA;
 
// file local variables declared, declared static so will not be visible in other files
volatile static byte irTempData[pfodIRTemp::IRTEMP_DATA_SIZE];
volatile static byte irTempDataBit;
volatile static byte irTempDataIndex;
volatile static bool irTempSensorEnabled;

/**
 * If the sensor is enabled then this interrupt is called on each falling edge
 * and saves the reading to irTempData until 5 bytes read
 * We know we are finished when irTempDataIndex == IRTEMP_DATA_SIZE
 * Note: irTempDataIndex is volatile so compiler won't optimize away checks of its value
 */
static void irTempInterrupt(void) {
  if (irTempSensorEnabled) {
    if (irTempDataIndex < pfodIRTemp::IRTEMP_DATA_SIZE) {
      irTempDataBit --;
      irTempData[irTempDataIndex] |= digitalRead(pfodIRTemp::PIN_DATA) << irTempDataBit;
      if (irTempDataBit == 0) {
        irTempDataBit = 8;
        irTempDataIndex ++;
      }
    }
  }
}

/**
 * Use this for Uno and other 'compatible' boards
 * setup I/O pins and attach interrupt to read input
 * This constructor uses 
 * D2 is the input data pin
 * Interrupt 1 on D3
 * D4 is the aquire trigger pin
 * 
 */
pfodIRTemp::pfodIRTemp(void ) {
   initIRTemp((byte)2,(byte)3,(byte)4,(byte)1);  // D2=input data pin, Interrupt 1 on D3 clock, D4 =aquire trigger pin, Interrupt 1 wired to D3 on Uno boards
}

/**
 * Use this for other non-Uno boards
 * For FioV3 use pfodIRTemp(D2,D3,D4,0); // D3 is connected to INT0 on this board
 * setup I/O pins and attach interrupt to read input
 * This constructor uses 
 * dataPin is the input data pin
 * clockPin is the input pin for the Sensor's clock output
 * aquirePin is the aquire trigger pin
 * clockInterrupt is the interrupt number that matches the clockPin
 * 
 */
pfodIRTemp::pfodIRTemp(byte dataPin, byte clockPin, byte aquirePin, byte clockInterrupt ) {
  initIRTemp(dataPin, clockPin, aquirePin, clockInterrupt);
}


void pfodIRTemp::initIRTemp(byte dataPin, byte clockPin, byte aquirePin, byte clockInterrupt ) { 
  PIN_DATA = dataPin;//    = 2;  for UNO/FioV3
  PIN_ACQUIRE = aquirePin;// = 4;  for UNO/FioV3

  pinMode(aquirePin, OUTPUT);
  pinMode(clockPin,   INPUT);
  pinMode(dataPin,    INPUT);

  digitalWrite(aquirePin, HIGH);
  digitalWrite(clockPin,   HIGH);
  digitalWrite(dataPin,    HIGH);
  sensorEnable(false);
  readingTimeout = 0;
  dataType = IRTEMP_DATA_IR; // hard coded for IR temp set to any other value to read ambient temp
  attachInterrupt(clockInterrupt, irTempInterrupt, FALLING);
}


/**
 * Call this each loop to see if reading has been received
 * Returns NO_DATA (0) if still waiting for reading
 * else returns reading in 16* Kelvin.
 * or < NO_DATA if data is invalid for any reason.
 *  INVALID_DATA (-1) if check sum error or
 *  READING_TIMEOUT (-2) if timed out waiting for reading
 *
 * dataType determines reading returned IRTEMP_DATA_IR returns IR temp,
 * any other value returns the ambient temp
 * hard coded in this library to always return IR temp
 */
int pfodIRTemp::getIRTemperature(void) {
  if (!irTempSensorEnabled) {
  	  return 0.0; // no reading
  }	  
  byte dataType = IRTEMP_DATA_IR; // hard coded for IR temp
  int temperature = INVALID_DATA; 
  unsigned long deltaMillis = 0; // clear last result
  unsigned long thisMillis = millis(); // do this just once to prevent getting different answers from multiple calls to millis()
  if (thisMillis != lastMillis) {
    // we have ticked over calculate how many millis have gone past
    deltaMillis = thisMillis-lastMillis;  // note this works even if millis() has rolled over back to 0
    lastMillis = thisMillis;
  }  
  
    if (irTempDataIndex >= IRTEMP_DATA_SIZE) {
    	// got full sample
		if (irTempData[0] == dataType) {
		  if (validData(irTempData)) temperature = readTemperature(irTempData);
		} else { 
		  if (validData(irTempData+5)) temperature = readTemperature(irTempData+5);
		}
	  sensorEnable(false); // next call to get getIRTemperature will return 0.0
	  return(temperature);  // if not valid then INVALID_DATA returned here
   } // else

   // still waiting see if we have timedout
   if (readingTimeout > 0) {
      readingTimeout -= deltaMillis; //count down 	   
	   if (readingTimeout <= 0) { // have timed out
		  sensorEnable(false); // next call to get getIRTemperature will return 0.0
		  return READING_TIMEOUT; // time out return == invalid
	   } else {
		 return NO_DATA;  // still waiting
	   }
   } //else 
   return NO_DATA;  // still waiting
	   
}

/**
 * Convert 16 * Kelvin reading to Centigrade
 */
float pfodIRTemp::convertToCelsius(int reading) {
   if (reading < 0) {
   	   // invalid 
   	   return -274.0;
   } // else 
   return (((float)reading) / 16.0 - 273.15);
}

/**
 * Convert 16 * Kelvin reading to Fahrenheit
 */
float pfodIRTemp::convertToFahrenheit(int reading) {
   // div by 16 multiply by 9/5
   // then add offset to get fahrenheit
   if (reading < 0) {
   	   // invalid 
   	   return -460.0;
   } // else 
   return (((float)reading) * 0.1125 - 459.67); // -273.15*(9/5) + 32
}

/**
 * Return reading in degrees Kelvin * 16
 */
int pfodIRTemp::readTemperature(volatile byte data[]) {
  int msb = data[1] << 8;
  int lsb = data[2];
  return(msb + lsb); // reading in deg Kelvin * 16
}

/**
 * reset data for next reading
 */
void pfodIRTemp::resetData(void) {
  for (byte i = 0;  i < IRTEMP_DATA_SIZE;  i ++) irTempData[i] = 0;
  irTempDataBit = 8;
  irTempDataIndex = 0;
}

/**
 * Trigger sensor to start taking reading
 * You can call this again and it will have no effect
 * until getIRtempurature returns a non-zero result
 * After getIRtempurature returns a non-zero result
 * the next call to triggerSensor will start another reading
 */
void pfodIRTemp::triggerSensor() {
      sensorEnable(true);
}

/**
 * Trigger sensor and initialize time out if state changes from false to true
 * ignore calls with state true if already true
 * Stop timer if state false
 */
void pfodIRTemp::sensorEnable(bool state) {
  if (state) {
     if (irTempSensorEnabled) {
  	  return; // already called
     } else {
       // !irTempSensorEnabled start new reading       
      resetData();
  	  lastMillis = millis(); // save this time point
      readingTimeout = IRTEMP_TIMEOUT; // max 2sec for next reading
    }  
  } else { // else if false just stop timer,  
    readingTimeout = 0; // stop timer
   }  
   // update irTempSensorEnabled and trigger
  irTempSensorEnabled = state;
  digitalWrite(PIN_ACQUIRE, ! irTempSensorEnabled);
}


/**
 * Check checksum of data to see if it is valid
 */
bool pfodIRTemp::validData(
  volatile byte data[]) {
  byte checksum = (data[0] + data[1] + data[2]) & 0xff;
  return(data[3] == checksum  &&  data[4] == 0x0d);
}

