/*
 * File:    pfodIRTemp.h
 * (c)2013 Forward Computing and Control Pty. Ltd
 * see www.pfod.com.au
 *
 * NOTE: No argument constructor uses Interrupt 1
 * and pins D2(data),D3(clock),D4(trigger)
 *
 * To start a measurement call 
 * triggerSensor() 
 * triggerSensor can be called every loop as subsequent calls do nothing until a non-reading has been returned by 
 * getIRtemperature();
 * 
 * Calls to getIRtemperature() will return pfodIRTemp.NO_DATA (0) until a reading is available
 * When a reading is available getIRtemperature() will return either the raw reading (16*Kelvin)
 * or < pfodIRTemp.NO_DATA if error (pfodIRTemp.INVALID_DATA (-1) or pfodIRTemp.READING_TIMEOUT (-2))
 *
 * Sample code
   #include "pfodIRTemp.h" 
   pfodIRTemp irTemp;  // construct class to handle sensor
   
   bool fahrenheit = false; // set to true for Fahrenheit instead of Celsius
   void setup() {
     Serial.begin(9600);
   }
   
   void loop() {
     irTemp.triggerSensor();
     int reading = irTemp.getIRTemperature();
     if (reading != pfodIRTemp::NO_DATA) {
       // have a reading
       if (reading <  pfodIRTemp::NO_DATA) {
          Serial.println(F(" ,Invalid Reading"));
       } else {
          if (fahrenheit) {
            Serial.print(irTemp.convertToFahrenheit(reading)); 
            Serial.println(F(" ,F"));
          } else {  
            Serial.print(irTemp.convertToCelsius(reading));
            Serial.println(F(" ,C"));
          }  
       }
     }  
   } 
 *
 */

#ifndef pfodIRTemp_h
#define pfodIRTemp_h
#include "Arduino.h"

class pfodIRTemp {
  public:
  	/**
  	 * Use this for Uno type boards
  	 */
    pfodIRTemp(); //sets int1 on D3 for clock-in, dataPin D2, aquirePin D4
    
	/**
	 * Use this for other non-Uno boards
	 * For FioV3 use pfodIRTemp(D2,D3,D4,0); // D3 is connected to INT0 on this board
	 * setup I/O pins and attach interrupt to read input
	 * This constructor uses 
	 * dataPin is the input data pin
	 * clockPin is the input pin for the Sensor's clock output
	 * aquirePin is the aquire trigger pin
	 * clockInterrupt is the interrupt number that matches the clockPin
	 */
    pfodIRTemp(byte dataPin, byte clockPin, byte aquirePin, byte clockInterrupt);
    void triggerSensor();  // trigger sensor to start taking reading.
    int getIRTemperature(void); // returns NO_DATA (0) (absolute zero) if no reading or not enabled since last reading
                                  // return < NO_DATA (ie. INVALID_DATA or READING_TIMEOUT if error, timeout or invalid data)
                                  // returns > NO_DATA for valid reading
    float convertToFahrenheit(int reading);
    float convertToCelsius(int reading);
    static const byte IRTEMP_DATA_SIZE; // = 10;
    static const int NO_DATA; // = 0;
    static const int INVALID_DATA;//    = -1;
    static const int READING_TIMEOUT;//   = -2;  
    static byte PIN_DATA;//    = 2;  for UNO/FioV3  need this public to access in interrupt method
  private:
  	void initIRTemp(byte dataPin, byte clockPin, byte aquirePin, byte clockInterrupt);
    void  resetData(void);
    void  sensorEnable(bool state);
    int readTemperature(volatile byte data[]);
    bool  validData(volatile byte data[]);
    int readingTimeout;
    volatile unsigned long lastMillis; 
    byte dataType;
    static const long IRTEMP_TIMEOUT;// = 1000;  // milliseconds reading timeout
    static const long IRTEMP_DATA_IR;// = 0x4C;
    static byte PIN_ACQUIRE;// = 4;  for UNO/FioV3
};
#endif
